/***************************************************************************
 *   Copyright (C) 2008 by Emil Ljungdahl <emil@kratern.se>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 *                                                                         *
 *   $Id: quicklauncher.cpp 14 2008-08-20 20:48:47Z emil $                  *
 ***************************************************************************/

#include "quicklauncher.h"

// KDE
#include <KIcon>
#include <kworkspace/kworkspace.h>


QuickLauncher::QuickLauncher(QObject *parent, const QVariantList &args)
	: 	Plasma::Applet(parent, args),
		add_dialog(0),
		conf_dialog(0),
		innerLayout(0),
		rowCount(2),
		visibleIcons(10),
		addAction(0),
		removeAction(0),
		rightClickedIcon(-1),
		showAllIcons(false)
{
	setHasConfigurationInterface(true);
	setAcceptDrops(true);
	setFlag(QGraphicsItem::ItemIsMovable, true);
}

void QuickLauncher::init()
{
	KConfigGroup cg = config();
	QStringList desktopFiles;
	desktopFiles = cg.readEntry("iconUrls", desktopFiles);
	int rows = cg.readEntry("rowCount", rowCount);
	if(rows > 0)
		rowCount = rows;
	int visIcons = cg.readEntry("visibleIcons", visibleIcons);
	if(visIcons > 0)
		visibleIcons = visIcons;

	setAspectRatioMode(Plasma::IgnoreAspectRatio);

	// Set outer layout
	QGraphicsLinearLayout *outerLayout = new QGraphicsLinearLayout(this);
	outerLayout->setContentsMargins(0, 0, 0, 0);
	outerLayout->setSpacing(0);
	setLayout(outerLayout);

	// Set inner layout
	innerLayout = new InnerLayout(0, rowCount);
	outerLayout->setContentsMargins(0, 0, 0, 0);
	innerLayout->setSpacing(0);
	outerLayout->addItem(innerLayout);

	// Create expand Icon
	arrowIcon = new Plasma::Icon(this);
	arrowIcon->setIcon(KIcon("arrow-right-double"));
	arrowIcon->setMaximumSize(QSizeF(18, 18));
	connect(arrowIcon, SIGNAL(clicked()), this, SLOT(toggleShowAllIcons()));
	// HACK to initalize innerLayout
	innerLayout->addItem(arrowIcon);
	innerLayout->removeAt(0);

	readDestopFiles(desktopFiles);
	fillWidget();

	updateLayout();
}

QuickLauncher::~QuickLauncher()
{
	if(add_dialog) delete add_dialog;
	if(conf_dialog) delete conf_dialog;
	if(addAction) delete addAction;
	if(removeAction) delete removeAction;
	for(int i=0; i<iconList.size(); i++) {
		MyIcon *icon = iconList[i];
		iconList.removeAt(i);
		delete icon;
	}
}

void QuickLauncher::createLayout()
{
	Plasma::Icon *icon;
	while((icon = dynamic_cast<Plasma::Icon*>(innerLayout->itemAt(0))) != 0) {
		innerLayout->removeAt(0);
		icon->hide();
	}
	innerLayout->setRowCount(rowCount);
	if(layout()->count() > 1)
		layout()->removeAt(1);
	arrowIcon->hide();
	if(visibleIcons < iconList.size()) {
		if(showAllIcons)
			arrowIcon->setIcon(KIcon("arrow-left-double"));
		else
			arrowIcon->setIcon(KIcon("arrow-right-double"));
		arrowIcon->show();
		QGraphicsLinearLayout *tmpLayout = dynamic_cast<QGraphicsLinearLayout*>(layout());
		if(tmpLayout != 0)
		{
			QGraphicsLinearLayout *newLayout = new QGraphicsLinearLayout(Qt::Vertical, tmpLayout);
			newLayout->addItem(arrowIcon);
			newLayout->addStretch();
			tmpLayout->addItem(newLayout);
		}
	}
}

void QuickLauncher::fillWidget()
{
	createLayout();
	int items = (showAllIcons) ? iconList.size() : qMin(visibleIcons, iconList.size());
	int cols = (int)(ceil(1.0*items/rowCount));
	int index = 0;
	for(int i=0; i<cols; i++) {
		for(int j=0; j<rowCount; j++, index++)
		{
			index = j*cols+i;
			if(index < items)
			{
				MyIcon *icon = iconList[index];
				icon->show();
				innerLayout->addItem(icon);
			}
		}
	}
	updateLayout();
}

void QuickLauncher::updateLayout()
{
	updateConstraints();
}

QSizeF QuickLauncher::contentSizeHint() const
{
	QSizeF sHint = QSizeF(64, 64);
	int items = (showAllIcons) ? iconList.size() : qMin(visibleIcons, iconList.size());
	double cols = ceil(1.0*items/rowCount);
	cols = (cols>0) ? cols : 1;

	if(formFactor() == Plasma::Planar ||
		formFactor() == Plasma::MediaCenter) {
		if (items != 0)
		{
			sHint = layout()->geometry().size();
			sHint.setWidth(sHint.height()/rowCount*cols+((visibleIcons < iconList.size()) ? 18 : 0));
		}
	}
	else
	{
		if (items != 0)
		{
			sHint = QSizeF(containment()->size().height(), containment()->size().height());
			sHint.setWidth(sHint.height()/rowCount*cols+((visibleIcons < iconList.size()) ? 18 : 0));
		}
	}
	return sHint;
}

void QuickLauncher::contextMenuEvent(QGraphicsSceneContextMenuEvent* event)
{
	rightClickedIcon = findClickedIcon(event->scenePos());
	Applet::contextMenuEvent(event);
}

bool QuickLauncher::sceneEventFilter(QGraphicsItem*, QEvent *event)
{
	switch (event->type()) {
		case QEvent::GraphicsSceneMousePress: {
			mousePressEvent(static_cast<QGraphicsSceneMouseEvent*>(event));
			break;
		}
		case QEvent::GraphicsSceneMouseMove: {
			mouseMoveEvent(static_cast<QGraphicsSceneMouseEvent*>(event));
			break;
		}

		default:
			break;
	}

	return false;
}

void QuickLauncher::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
	setFlag(QGraphicsItem::ItemIsMovable, true);
	if(event->button() == Qt::LeftButton) {
		startPos = event->scenePos();
		rightClickedIcon = findClickedIcon(startPos);
	}
	Plasma::Applet::mousePressEvent(event);
}

void QuickLauncher::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{
	if(event->buttons() & Qt::LeftButton) {
		int distance = (event->scenePos() - startPos).toPoint().manhattanLength();
		if(distance >= QApplication::startDragDistance())
			startDrag(event);
	}
}

void QuickLauncher::dragEnterEvent(QGraphicsSceneDragDropEvent *event)
{
	event->setDropAction(Qt::MoveAction);
	event->accept();
}

void QuickLauncher::dragMoveEvent(QGraphicsSceneDragDropEvent* event)
{
	event->setDropAction(Qt::MoveAction);
	event->accept();
}

void QuickLauncher::dropEvent(QGraphicsSceneDragDropEvent* event)
{
	// Calculate position
	QPointF point = event->pos();
	int items = (showAllIcons) ? iconList.size() : qMin(visibleIcons, iconList.size());
	int cols = static_cast<int>(ceil(1.0*items/rowCount));
	int col = static_cast<int>(round(point.x()*cols/innerLayout->geometry().width()));
	col = (col >= cols) ? col-1 : col;
	int row = static_cast<int>(floor(point.y()*rowCount/size().height()));
	row = (row >= rowCount) ? row-1 : row;
	int pos = row*cols+col;

	if(pos >= items)
		pos = items-1;

	if(rightClickedIcon != -1) {
		iconList.move(rightClickedIcon, pos);
		rightClickedIcon = -1;
	} else {
		if (!KUrl::List::canDecode(event->mimeData())) {
			return;
		}

		//FIXME: Should use some built in method...
		QString payload = event->mimeData()->text();
		if (payload.isEmpty()) {
			return;
		}
		KUrl::List urls(payload.split("\n"));

		//if there are more than one the last is junk
		if (urls.count() > 1) {
			urls.removeLast();
		}
	//     KUrl::List urls = KUrl::List::fromMimeData(event->mimeData());

		if (urls.count() > 0) {
			event->setDropAction(Qt::MoveAction);
			event->accept();
		} else {
			return;
		}

		KConfigGroup cg = config();

		foreach (KUrl url, urls) {
			if(KDesktopFile::isDesktopFile(url.toLocalFile())) {
				readDesktopFile(pos, url.toLocalFile());
			}
		}
	}
	saveConfig();
	fillWidget();
}

QList<QAction*> QuickLauncher::contextualActions()
{
	QList<QAction*> tempActions;
	if (addAction == NULL) {
		addAction = new QAction(KIcon("list-add"), i18n("Add Icon..."), this);
		connect(addAction, SIGNAL(triggered(bool)), this, SLOT(showAddInterface()));
	}
	tempActions << addAction;

	if(rightClickedIcon > -1) {
		if(removeAction == NULL) {
			removeAction = new QAction(KIcon("list-remove"), i18n("Remove Icon..."), this);
			connect(removeAction, SIGNAL(triggered(bool)), this, SLOT(removeIcon()));
		}
		tempActions << removeAction;
	}

	return tempActions;
}

void QuickLauncher::showConfigurationInterface()
{
	if (conf_dialog == 0) {
		conf_dialog = new KDialog;
		conf_dialog->setCaption(i18n("Configure QuickLauncher"));

		QWidget *widget = new QWidget;
		uiConfig.setupUi(widget);
		conf_dialog->setMainWidget(widget);
		conf_dialog->setButtons( KDialog::Ok | KDialog::Cancel | KDialog::Apply );

		connect( conf_dialog, SIGNAL(applyClicked()), this, SLOT(configAccepted()) );
		connect( conf_dialog, SIGNAL(okClicked()), this, SLOT(configAccepted()) );

		uiConfig.rowCount->setValue(rowCount);
		uiConfig.visibleIcons->setValue(visibleIcons);

	}
	conf_dialog->show();
}

void QuickLauncher::showAddInterface()
{
	if (add_dialog == 0) {
		add_dialog = new KDialog;
		add_dialog->setCaption(i18n("Add shortcut"));

		QWidget *widget = new QWidget;
		uiAdd.setupUi(widget);
		add_dialog->setMainWidget(widget);
		add_dialog->setButtons( KDialog::Ok | KDialog::Cancel );

		connect( add_dialog, SIGNAL(okClicked()), this, SLOT(addAccepted()) );

	}
	add_dialog->show();
}

void QuickLauncher::removeIcon()
{
	removeIcon(rightClickedIcon);
	rightClickedIcon = -1;
}

void QuickLauncher::toggleShowAllIcons()
{
	showAllIcons = !showAllIcons;
	fillWidget();
}

void QuickLauncher::hideIcons()
{
	if(showAllIcons) {
		showAllIcons = false;
		fillWidget();
	}
}

void QuickLauncher::addAccepted()
{
	int insertplace = (rightClickedIcon>-1) ? rightClickedIcon : iconList.size();
	readDesktopFile(insertplace, uiAdd.urlIcon->url().url());
	fillWidget();
}

void QuickLauncher::configAccepted()
{
	KConfigGroup cg = config();
	int rows = uiConfig.rowCount->value();
	int visIcons = uiConfig.visibleIcons->value();
	if(rows > 0 && rows != rowCount)
	{
		cg.writeEntry("rowCount", rows);
		rowCount = rows;
	}
	if(visIcons > 0 && visIcons != visibleIcons)
	{
		cg.writeEntry("visibleIcons", visIcons);
		visibleIcons = visIcons;
	}
	emit configNeedsSaving();
	fillWidget();
	updateConstraints(Plasma::SizeConstraint);
}

void QuickLauncher::constraintsEvent(Plasma::Constraints constraints)
{
	if (constraints & Plasma::SizeConstraint || constraints & Plasma::FormFactorConstraint) {
		if(formFactor() == Plasma::Planar ||
		   formFactor() == Plasma::MediaCenter) {
			setMinimumSize(48, 48);
			setPreferredSize(contentSizeHint());

			QSizeF sHint = layout()->geometry().size();
			for(int i=0; i<iconList.size(); i++)
			{
				Plasma::Icon *icon = iconList[i];
				icon->resize(QSizeF((sHint.height()-sHint.height()*0.06)/rowCount, (sHint.height()-sHint.height()*0.06)/rowCount));
			}
		}
		else
		{
			setBackgroundHints(NoBackground);
			setMinimumSize(0, 0);
			setPreferredSize(contentSizeHint());
			layout()->setPreferredSize(contentSizeHint());

			QSizeF sHint = QSizeF(containment()->size().height(), containment()->size().height());
			for(int i=0; i<iconList.size(); i++)
			{
				Plasma::Icon *icon = iconList[i];
				icon->resize(QSizeF((sHint.height()-sHint.height()*0.06)/rowCount, (sHint.height()-sHint.height()*0.06)/rowCount));
			}
		}
		innerLayout->updateGeometry();
	}
}

void QuickLauncher::readDesktopFile(int index, QString url)
{
	KUrl appUrl = KUrl(url);
	QString mimetype = KMimeType::findByUrl(appUrl)->name();

	MyIcon *icon = new MyIcon(mimetype, appUrl, this);
	icon->installSceneEventFilter(this);

	if (appUrl.isLocalFile() && KDesktopFile::isDesktopFile(appUrl.toLocalFile())) {
		KDesktopFile *f= new KDesktopFile(appUrl.toLocalFile());
		icon->setIcon(f->readIcon());
	} else {
		icon->setIcon(KMimeType::iconNameForUrl(appUrl));
	}

	if (icon->icon().isNull()) {
		icon->setIcon("unknown");
	}
	icon->hide();
	connect(icon, SIGNAL(clicked()), this, SLOT(hideIcons()));
	iconList.insert(index, icon);
}

void QuickLauncher::readDestopFiles(QStringList desktopFiles)
{
    // Loop through list
    for(int i=0; i<desktopFiles.size(); i++)
        readDesktopFile(i, desktopFiles[i]);
}

void QuickLauncher::removeIcon(int iconIndex)
{
	if(iconIndex > -1 && iconIndex < iconList.size())
	{
		MyIcon *icon = iconList[iconIndex];
		iconList.removeAt(iconIndex);

		fillWidget();
		delete icon;

		saveConfig();
	}
}

int QuickLauncher::findClickedIcon(QPointF point)
{
	QGraphicsItem* item = scene()->itemAt(point);
	Plasma::Icon *icon = 0;
	if(item) {
		icon = qgraphicsitem_cast<Plasma::Icon*>(item);
	}
	if(icon) {
		for(int i=0; rightClickedIcon <= -1 && i < iconList.size(); i++) {
			if(icon == iconList[i])
				return i;
		}
	}
	return -1;
}

void QuickLauncher::startDrag(QGraphicsSceneMouseEvent *event)
{
	if(rightClickedIcon > -1) {
		QMimeData *mimeData = new QMimeData;
		KUrl::List urls(iconList[rightClickedIcon]->getUrl());
		urls.populateMimeData(mimeData);
		QDrag *drag = new QDrag(event->widget());
		drag->setMimeData(mimeData);
		drag->start(Qt::MoveAction);
	}
}

void QuickLauncher::saveConfig()
{
	KConfigGroup cg = config();
	QStringList desktopFiles;
	for(int i=0; i<iconList.size(); i++)
		desktopFiles.append(iconList[i]->getPath());
	cg.writeEntry("iconUrls", desktopFiles);
	emit configNeedsSaving();
}

#include "quicklauncher.moc"
